/*
 * Copyright (C) 2012 One Laptop per Child
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Written by: Daniel Drake
 *
 */

//#define PLY_ENABLE_TRACING

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>

#include <ply-boot-splash-plugin.h>
#include <ply-image.h>
#include <ply-pixel-buffer.h>
#include <ply-logger.h>

struct _ply_boot_splash_plugin
{
	ply_event_loop_t *loop;
	ply_boot_splash_mode_t mode;

	ply_image_t *xo_image;
	ply_image_t *dot_image;
	ply_image_t *dot_light_image;
	ply_image_t *sl_image;
	ply_image_t *fedora_image;
	ply_image_t *shutdown_image;
	ply_image_t *custom_image;

	/* Number of the last frame (0-3) shown.
	 * Frame 0 shows the dots 0, 4, 8, 12, 16, 20 as dark, the rest as light.
	 * Frame 1 shows the dots 1, 5, 9, 13, 17, 21 as dark, the rest as light.
	 * Frame 2 shows the dots 2, 6, 10, 14, 18, 22 as dark, the rest as light.
	 * Frame 3 shows the dots 3, 7, 11, 15, 19, 23 as dark, the rest as light.
	 */
	int frame;

	bool finished;
	bool has_custom_image;

	/* Only draw on one display, to keep things simple. */
	ply_pixel_display_t *display;
};

/* Offset-from-center coordinates of the 24 dots that form a circle around the
 * XO icon. Each coordinate specifies the top left point of the 22x22 dot. */
static const struct coord {
	int x;
	int y;
} dot_coords[] = {
	{ -11, -211 }, /* 12 o'clock */
	{ 41, -204 },
	{ 89, -185 },
	{ 130, -152 },
	{ 162, -111 },
	{ 182, -63 },
	{ 189, -11 }, /* 3 o'clock */
	{ 182, 41 },
	{ 162, 89 },
	{ 130, 130 },
	{ 89, 162 },
	{ 41, 182 },
	{ -11, 189 }, /* 6 o'clock */
	{ -63, 182 },
	{ -111, 162 },
	{ -152, 130 },
	{ -184, 89 },
	{ -204, 41 },
	{ -211, -11 }, /* 9 o'clock */
	{ -204, -63 },
	{ -184, -111 },
	{ -152, -152 },
	{ -111, -185 },
	{ -63, -204 },
};
#define NR_DOTS 24

/* offset into above array of low-central "6 o'clock" point */
#define SIXOCLOCK 12

static void invalidate_dot_area(ply_pixel_display_t *display, int dotnr,
								unsigned long screen_width,
								unsigned long screen_height)
{
	const struct coord *coord = &dot_coords[dotnr];
	int x = (screen_width / 2) + coord->x;
	int y = (screen_height / 2) + coord->y;
	ply_pixel_display_draw_area(display, x, y, 22, 22);
}

static void
draw_shutdown(ply_boot_splash_plugin_t *plugin,
			  ply_pixel_buffer_t *pixel_buffer)
{
	ply_rectangle_t image_area;
	uint32_t *img_data;

	img_data = ply_image_get_data(plugin->shutdown_image);
	image_area.width = ply_image_get_width(plugin->shutdown_image);
	image_area.height = ply_image_get_height(plugin->shutdown_image);
	image_area.x = 0;
	image_area.y = 0;
	ply_pixel_buffer_fill_with_argb32_data(pixel_buffer, &image_area, img_data);
}


static void
on_draw(ply_boot_splash_plugin_t *plugin, ply_pixel_buffer_t *pixel_buffer,
		int	x, int y, int width, int height)
{
	ply_rectangle_t screen_area;
	ply_rectangle_t image_area;
	uint32_t *img_data;
	uint32_t *dot_data;
	uint32_t *dot_light_data;
	int i;

	if (plugin->mode == PLY_BOOT_SPLASH_MODE_SHUTDOWN) {
		draw_shutdown(plugin, pixel_buffer);
		return;
	}

	ply_pixel_buffer_get_size(pixel_buffer, &screen_area);
	ply_pixel_buffer_fill_with_hex_color(pixel_buffer, &screen_area, ~0);

	/* draw XO man */
	img_data = ply_image_get_data(plugin->xo_image);
	image_area.width = ply_image_get_width(plugin->xo_image);
	image_area.height = ply_image_get_height(plugin->xo_image);
	image_area.x = screen_area.width / 2 - image_area.width / 2;
	image_area.y = screen_area.height / 2 - image_area.height / 2;
	ply_pixel_buffer_fill_with_argb32_data(pixel_buffer, &image_area, img_data);

	/* draw SugarLabs logo */
	img_data = ply_image_get_data(plugin->sl_image);
	image_area.width = ply_image_get_width(plugin->sl_image);
	image_area.height = ply_image_get_height(plugin->sl_image);
	image_area.x = 0;
	image_area.y = screen_area.height - image_area.height;
	ply_pixel_buffer_fill_with_argb32_data(pixel_buffer, &image_area, img_data);

	/* draw Fedora logo */
	img_data = ply_image_get_data(plugin->fedora_image);
	image_area.width = ply_image_get_width(plugin->fedora_image);
	image_area.height = ply_image_get_height(plugin->fedora_image);
	image_area.x = screen_area.width - image_area.width;
	image_area.y = screen_area.height - image_area.height;
	ply_pixel_buffer_fill_with_argb32_data(pixel_buffer, &image_area, img_data);

	/* draw custom image */
	if (plugin->has_custom_image) {
		img_data = ply_image_get_data(plugin->custom_image);
		image_area.width = ply_image_get_width(plugin->custom_image);
		image_area.height = ply_image_get_height(plugin->custom_image);
		image_area.x = screen_area.width / 2 - image_area.width / 2;
		image_area.y = screen_area.height - image_area.height;
		ply_pixel_buffer_fill_with_argb32_data(pixel_buffer, &image_area, img_data);
	}

	/* draw dots */
	dot_data = ply_image_get_data(plugin->dot_image);
	dot_light_data = ply_image_get_data(plugin->dot_light_image);
	image_area.width = 22;
	image_area.height = 22;
	for (i = 0; i < NR_DOTS; i++) {
		uint32_t *data = dot_light_data;
		if (plugin->finished || (i - plugin->frame) % 4 == 0)
			data = dot_data;
		image_area.x = screen_area.width / 2 + dot_coords[i].x;
		image_area.y = screen_area.height / 2 + dot_coords[i].y;
		ply_pixel_buffer_fill_with_argb32_data(pixel_buffer, &image_area, data);
	}
}

static void
redraw(ply_boot_splash_plugin_t *plugin)
{
	unsigned long screen_width, screen_height;
	ply_pixel_display_t *display = plugin->display;

	if (!display)
		return;

	screen_width = ply_pixel_display_get_width(display);
	screen_height = ply_pixel_display_get_height(display);
	ply_pixel_display_draw_area(display, 0, 0, screen_width, screen_height);
}

static void
add_pixel_display(ply_boot_splash_plugin_t *plugin,
					ply_pixel_display_t *display)
{
	if (plugin->display == NULL) {
		ply_trace("adding pixel display");
		plugin->display = display;
		ply_pixel_display_set_draw_handler(display,
										   (ply_pixel_display_draw_handler_t) on_draw,
										    plugin);
	} else {
		ply_trace("ignoring multiple pixel displays");
	}
}

static void
remove_pixel_display(ply_boot_splash_plugin_t *plugin,
						ply_pixel_display_t *display)
{
	if (plugin->display == display) {
		plugin->display = NULL;
		ply_pixel_display_set_draw_handler(display, NULL, NULL);
		ply_trace("removed pixel display");
	}
}

static void
on_timeout(ply_boot_splash_plugin_t *plugin)
{
	ply_pixel_display_t *display = plugin->display;
	unsigned long screen_width = ply_pixel_display_get_width(display);
	unsigned long screen_height = ply_pixel_display_get_height(display);
	int i;
	int old_frame = plugin->frame;

	ply_trace("on_timeout current frame %d", plugin->frame);

	/* move to next frame */
	plugin->frame++;
	if (plugin->frame == 4)
		plugin->frame = 0;

	/* invalidate dark dots in old frame
	 * (so that they get repainted as light) */
	for (i = old_frame; i < NR_DOTS; i += 4)
		invalidate_dot_area(display, i, screen_width, screen_height);

	/* invalidate dark dots in new frame
	 * (so that currently light ones get repainted as dark) */
	for (i = plugin->frame; i < NR_DOTS; i += 4)
		invalidate_dot_area(display, i, screen_width, screen_height);

	/* reschedule */
	ply_event_loop_watch_for_timeout(plugin->loop, 1.0,
									 (ply_event_loop_timeout_handler_t) on_timeout,
									 plugin);	
}

static bool
show_splash_screen(ply_boot_splash_plugin_t *plugin, ply_event_loop_t *loop,
				   ply_buffer_t *boot_buffer, ply_boot_splash_mode_t mode)
{
	plugin->loop = loop;
	plugin->mode = mode;

	if (mode == PLY_BOOT_SPLASH_MODE_BOOT_UP) {
		ply_trace("loading images");
		plugin->has_custom_image = ply_image_load(plugin->custom_image);
		if (!ply_image_load(plugin->xo_image))
			return false;
		if (!ply_image_load(plugin->dot_image))
			return false;
		if (!ply_image_load(plugin->dot_light_image))
			return false;
		if (!ply_image_load(plugin->sl_image))
			return false;
		if (!ply_image_load(plugin->fedora_image))
			return false;

		plugin->frame = 0;
		ply_event_loop_watch_for_timeout(loop, 1.0,
										 (ply_event_loop_timeout_handler_t) on_timeout,
										 plugin);
	} else {
		if (!ply_image_load(plugin->shutdown_image))
			return false;
	}
	redraw(plugin);
	return true;
}

static void
update_status (ply_boot_splash_plugin_t *plugin, const char *status)
{
	if (strcmp(status, "finish") == 0) {
		plugin->finished = true;
		ply_event_loop_stop_watching_for_timeout(plugin->loop,
												 (ply_event_loop_timeout_handler_t) on_timeout,
												 plugin);
		redraw(plugin);
	}
}

static void
hide_splash_screen(ply_boot_splash_plugin_t *plugin, ply_event_loop_t *loop)
{
	ply_event_loop_stop_watching_for_timeout(loop,
											 (ply_event_loop_timeout_handler_t) on_timeout,
											 plugin);
}

static ply_boot_splash_plugin_t *
create_plugin(ply_key_file_t *key_file)
{
	ply_boot_splash_plugin_t *plugin;
	char *image_dir, *image_path;

	plugin = calloc(1, sizeof (ply_boot_splash_plugin_t));

	image_dir = ply_key_file_get_value (key_file, "olpc", "ImageDir");
	ply_trace ("Using '%s' as working directory", image_dir);

	asprintf(&image_path, "%s/xo.png", image_dir);
	plugin->xo_image = ply_image_new(image_path);
	free(image_path);

	asprintf(&image_path, "%s/dot.png", image_dir);
	plugin->dot_image = ply_image_new(image_path);
	free(image_path);

	asprintf(&image_path, "%s/dot_light.png", image_dir);
	plugin->dot_light_image = ply_image_new(image_path);
	free(image_path);

	asprintf(&image_path, "%s/sugarlabs.png", image_dir);
	plugin->sl_image = ply_image_new(image_path);
	free(image_path);

	asprintf(&image_path, "%s/fedora.png", image_dir);
	plugin->fedora_image = ply_image_new(image_path);
	free(image_path);

	asprintf(&image_path, "%s/shutdown.png", image_dir);
	plugin->shutdown_image = ply_image_new(image_path);
	free(image_path);

	asprintf(&image_path, "%s/custom.png", image_dir);
	plugin->custom_image = ply_image_new(image_path);
	free(image_path);

	return plugin;
}

static void
destroy_plugin(ply_boot_splash_plugin_t *plugin)
{
	if (plugin == NULL)
		return;

	ply_trace("destroying plugin");
	ply_image_free(plugin->xo_image);
	ply_image_free(plugin->dot_image);
	ply_image_free(plugin->dot_light_image);
	ply_image_free(plugin->sl_image);
	ply_image_free(plugin->fedora_image);
	free(plugin);
}

__attribute__((visibility("default")))
ply_boot_splash_plugin_interface_t *
ply_boot_splash_plugin_get_interface(void)
{
	static ply_boot_splash_plugin_interface_t plugin_interface = {
		.create_plugin = create_plugin,
		.destroy_plugin = destroy_plugin,
		.add_pixel_display = add_pixel_display,
		.remove_pixel_display = remove_pixel_display,
		.show_splash_screen = show_splash_screen,
		.hide_splash_screen = hide_splash_screen,
		.update_status = update_status,
	};

	return &plugin_interface;
}
